{ A Test-Program for the unit unextfloat.pas

  Version Date        Change
  0.0.1   2024-09-05  First release for Doubles
  0.0.2   2024-09-06  Implementation of the test for Singles
}

unit unextfloattestmain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, ComCtrls,
  ExtCtrls, Math,
  unextfloat, umsvcrtdll;

type

  { TForm1 }

  TForm1 = class(TForm)
    btnStart: TButton;
    Label1: TLabel;
    lblResult: TLabel;
    Memo1: TMemo;
    Panel1: TPanel;
    pbProgress: TProgressBar;
    procedure btnStartClick(Sender: TObject);
  private
    function TestFromNegToPosDouble(const AStringList : TStrings) : Boolean;
    function TestFromPosToNegDouble(const AStringList : TStrings) : Boolean;
    function TestFromNegToPosSingle(const AStringList : TStrings) : Boolean;
    function TestFromPosToNegSingle(const AStringList : TStrings) : Boolean;

  public

  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }

procedure TForm1.btnStartClick(Sender: TObject);
var
  sl : TStringList;
  b : Boolean;
begin
  lblResult.Caption := 'Running...';
  lblResult.Repaint;
  sl := TStringList.Create;
  try
    sl.Add('Begin of test');
    sl.Add('Test 1 of 4: Double Running from -Infinity to MaxDouble');
    b := False;
    if TestFromNegToPosDouble(sl) then
    begin
      sl.Add('Test 1 of 4: Double Finished Running from -Infinity to MaxDouble');
      sl.Add('Test 2 of 4: Double Running from Infinity to -MaxDouble');
      if TestFromPosToNegDouble(sl) then
      begin
        lblResult.Caption := 'Success';
        sl.Add('Test 2 of 4: Double Finished Running from Infinity to -MaxDouble');

        sl.Add('Test 3 of 4: Single Running from -Infinity to MaxSingle');
        if TestFromNegToPosSingle(sl) then
        begin
          sl.Add('Test 3 of 4: Single Finished Running from -Infinity to MaxSingle');
          sl.Add('Test 4 of 4: Single Running from Infinity to -MaxDouble');
          if TestFromPosToNegSingle(sl) then
          begin
            sl.Add('Test 4 of 4: Single Finished Running from Infinity to -MaxSingle');
            b := True;
          end;
        end;
      end;
    end;
    if b then
      lblResult.Caption := 'Succeeded'
    else
      lblResult.Caption := 'Failed';

    Memo1.Lines.BeginUpdate;
    try
      sl.Add('End of tests');
      Memo1.Lines.Assign(sl);
    finally
      Memo1.Lines.EndUpdate;
    end;
    pbProgress.Position := pbProgress.Max;
  finally
    sl.Free;
  end;
end;
(*

// Button2 was an internal Test for the euality tests of MaxSingle and -MaxSingle
// The problem is solved, but the code might be helpful for later examination
procedure TForm1.Button2Click(Sender: TObject);
var
  smax, s0, s1 : Single;
begin
  s0 := SingleMaxSingle;
  s1 := SingleMaxSingle;
  Memo1.Lines.Add(Format('MaxSingle %8.8x, s0 %8.8x, s1 %8.8x ',
                        [SingleToLongBits(SingleMaxSingle), SingleToLongBits(s0), SingleToLongBits(s1)]));
  if s0 = s1 then
    Memo1.Lines.Add('Equal (0)')
  else
    Memo1.Lines.Add('Not equal (0)');
  if s0 = SingleMaxSingle then
    Memo1.Lines.Add('Equal (1)')
  else
    Memo1.Lines.Add('Not equal (1)');
  if s1 = SingleMaxSingle then
    Memo1.Lines.Add('Equal (2)')
  else
    Memo1.Lines.Add('Not equal (2)');
  if SingleToLongBits(s0) = SingleToLongBits(s1) then
    Memo1.Lines.Add('Equal (3)')
  else
    Memo1.Lines.Add('Not equal (3)');
  if SingleToLongBits(s0) = SingleToLongBits(SingleMaxSingle) then
    Memo1.Lines.Add('Equal (4)')
  else
    Memo1.Lines.Add('Not equal (4)');
  if SingleToLongBits(s1) = SingleToLongBits(SingleMaxSingle) then
    Memo1.Lines.Add('Equal (4)')
  else
    Memo1.Lines.Add('Not equal (4)');
end;
*)
function TForm1.TestFromNegToPosDouble(const AStringList: TStrings): Boolean;
const
  ExpectedTests = 4283;
  loops = 1000;
var
  ddll, dfpc : Double;
  ds, dd : Double;
  i : Integer;
  cnt : Integer;
  d : Double;
  s : String;

begin
  Result := False;
  cnt := 0;
  dd := MaxDouble;
  ds := -Infinity;
  repeat
    d := cnt / ExpectedTests;
    if d > 1.0 then
      d := 1.0;
    pbProgress.Position := Trunc(d* 100.0);
    if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
      s := Format('[%5.5d] Start %1.30f',[cnt, ds])
    else
      s := Format('[%5.5d] Start %1.30g',[cnt, ds]);


    for i := 0 to loops-1 do
    begin
      if ds = 0.0 then
        s := s + ' (Testing 0.0) ';
      ddll := umsvcrtdll.nextafter(ds,dd);
      dfpc := unextfloat.NextAfter(ds,dd);
      if ddll <> dfpc then
      begin
        if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
          s := s + Format(' Stop %1.30f',[ds])
        else
          s := s + Format(' Stop %1.30g',[ds]);
        AStringList.Add(s);
        AStringList.Add('Test Failed!');
        Exit;
      end;
      if dfpc = MaxDouble then Break;
      ds := dfpc;
    end;
    if dfpc = MaxDouble then
    begin
{
      // This failes with umsvcrtdll.nextafter, causing a floating point violation
      ddll := umsvcrtdll.nextafter(MaxDouble,Infinity);
      // This succeeds with returning infinity
      dfpc := unextfloat.NextAfter(MaxDouble,Infinity);
}
      s := s + Format(' Stop %1.30g (MaxDouble)',[ds]);
      AStringList.Add(s);
      AStringList.Add('Test succeeded');
      Result := True;
      Break;
    end;
    if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
      s := s + Format(' Stop %1.30f',[ds])
    else
      s := s + Format(' Stop %1.30g',[ds]);
    AStringList.Add(s);

    if ds > 0 then
    begin
      if ds > MaxDouble / 2.0 then
        ds := ds + ((MaxDouble - ds) / 2.0)
      else
        ds := ds * 2.0;
    end
    else
    begin
      if ds < -MaxDouble / 2.0 then
        ds := ds + ((MaxDouble - Abs(ds)) / 2.0)
      else
        ds := ds / 2.0;
    end;
    Inc(cnt)
  until False;
end;

function TForm1.TestFromPosToNegDouble(const AStringList: TStrings): Boolean;
const
  ExpectedTests = 4283;
  loops = 1000;
var
  ddll, dfpc : Double;
  ds, dd : Double;
  i : Integer;
  cnt : Integer;
  d : Double;
  s : String;

begin
  Result := False;
  cnt := 0;
  dd := -MaxDouble;
  ds := Infinity;
  repeat
    d := cnt / ExpectedTests;
    if d > 1.0 then
      d := 1.0;
    pbProgress.Position := Trunc(d* 100.0);
    if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
      s := Format('[%5.5d] Start %1.30f',[cnt, ds])
    else
      s := Format('[%5.5d] Start %1.30g',[cnt, ds]);


    for i := 0 to loops-1 do
    begin
      if ds = 0.0 then
        s := s + ' (Testing 0.0) ';
      ddll := umsvcrtdll.nextafter(ds,dd);
      dfpc := unextfloat.NextAfter(ds,dd);
      if ddll <> dfpc then
      begin
        if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
          s := s + Format(' Stop %1.30f',[ds])
        else
          s := s + Format(' Stop %1.30g',[ds]);
        AStringList.Add(s);
        AStringList.Add('Test Failed!');
        Exit;
      end;
      if dfpc = -MaxDouble then Break;
      ds := dfpc;
    end;
    if dfpc = -MaxDouble then
    begin

{
      // This failes with umsvcrtdll.nextafter, causing a floating point violation
      ddll := umsvcrtdll.nextafter(-MaxDouble,-Infinity);
      // This succeeds with returning -Infinity
      dfpc := unextfloat.NextAfter(-MaxDouble,-Infinity);
}
      s := s + Format(' Stop %1.30g (MaxDouble)',[ds]);
      AStringList.Add(s);
      AStringList.Add('Test succeeded');
      Result := True;
      Break;
    end;
    if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
      s := s + Format(' Stop %1.30f',[ds])
    else
      s := s + Format(' Stop %1.30g',[ds]);
    AStringList.Add(s);

    if ds > 0 then
    begin
      if ds > MaxDouble / 2.0 then
        ds := ds - ((MaxDouble - ds) / 2.0)
      else
        ds := ds / 2.0;
    end
    else
    begin
      if ds < -MaxDouble / 2.0 then
        ds := ds - ((MaxDouble - Abs(ds)) / 2.0)
      else
        ds := ds * 2.0;
    end;
    Inc(cnt)
  until False;
end;

function TForm1.TestFromNegToPosSingle(const AStringList: TStrings): Boolean;
const
  ExpectedTests = 563;
  loops = 1000;
var
  fdll, ffpc : Single;
  ds, dd : Single;
  i : Integer;
  cnt : Integer;
  d : Double;
  s : String;
begin
  Result := False;
  cnt := 0;
  dd := MaxSingle;
  ds := -Infinity;
  repeat
    d := cnt / ExpectedTests;
    if d > 1.0 then
      d := 1.0;
    pbProgress.Position := Trunc(d* 100.0);
    if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
      s := Format('[%5.5d] Start %1.30f',[cnt, ds])
    else
      s := Format('[%5.5d] Start %1.30g',[cnt, ds]);


    for i := 0 to loops-1 do
    begin
      if ds = 0.0 then
        s := s + ' (Testing 0.0) ';
      fdll := umsvcrtdll.nextafterf(ds,dd);
      ffpc := unextfloat.NextAfter(ds,dd);
      if fdll <> ffpc then
      begin
        if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
          s := s + Format(' Stop %1.30f',[ds])
        else
          s := s + Format(' Stop %1.30g',[ds]);
        AStringList.Add(s);
        AStringList.Add('Test Failed!');
        Exit;
      end;
      if ffpc = SingleMaxSingle then Break;
      ds := ffpc;
    end;
    if ffpc = SingleMaxSingle then
    begin
{
      // This failes with umsvcrtdll.nextafter, causing a floating point violation
      fdll := umsvcrtdll.nextafterf(MaxSingle,Infinity);
      // This succeeds with returning infinity
      ffpc := unextfloat.NextAfter(MaxSingle,Infinity);
}
      s := s + Format(' Stop %1.30g (MaxSingle)',[ds]);
      AStringList.Add(s);
      AStringList.Add('Test succeeded');
      Result := True;
      Break;
    end;
    if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
      s := s + Format(' Stop %1.30f',[ds])
    else
      s := s + Format(' Stop %1.30g',[ds]);
    AStringList.Add(s);

    if ds > 0 then
    begin
      if ds > MaxSingle / 2.0 then
        ds := ds + ((MaxSingle - ds) / 2.0)
      else
        ds := ds * 2.0;
    end
    else
    begin
      if ds < -MaxSingle / 2.0 then
        ds := ds + ((MaxSingle - Abs(ds)) / 2.0)
      else
        ds := ds / 2.0;
    end;
    Inc(cnt)
  until False;
end;

function TForm1.TestFromPosToNegSingle(const AStringList: TStrings): Boolean;
const
  ExpectedTests = 563;
  loops = 1000;
var
  fdll, ffpc : Single;
  ds, dd : Single;
  i : Integer;
  cnt : Integer;
  d : Double;
  s : String;
begin
  Result := False;
  cnt := 0;
  dd := -MaxSingle;
  ds := Infinity;
  repeat
    d := cnt / ExpectedTests;
    if d > 1.0 then
      d := 1.0;
    pbProgress.Position := Trunc(d* 100.0);
    if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
      s := Format('[%5.5d] Start %1.30f',[cnt, ds])
    else
      s := Format('[%5.5d] Start %1.30g',[cnt, ds]);


    for i := 0 to loops-1 do
    begin
      if ds = 0.0 then
        s := s + ' (Testing 0.0) ';
      fdll := umsvcrtdll.nextafterf(ds,dd);
      ffpc := unextfloat.NextAfter(ds,dd);
      if fdll <> ffpc then
      begin
        if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
          s := s + Format(' Stop %1.30f',[ds])
        else
          s := s + Format(' Stop %1.30g',[ds]);
        AStringList.Add(s);
        AStringList.Add('Test Failed!');
        Exit;
      end;
      if ffpc = NegSingleMaxSingle then Break;
      ds := ffpc;
    end;
    if ffpc = NegSingleMaxSingle then
    begin
{
      // This failes with umsvcrtdll.nextafter, causing a floating point violation
      fdll := umsvcrtdll.nextafterf(NegSingleMaxSingle,-Infinity);
      // This succeeds with returning -Infinity
      ffpc := unextfloat.NextAfter(NegSingleMaxSingle,Single(-Infinity));
}
      s := s + Format(' Stop %1.30g (MaxSingle)',[ds]);
      AStringList.Add(s);
      AStringList.Add('Test succeeded');
      Result := True;
      Break;
    end;
    if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
      s := s + Format(' Stop %1.30f',[ds])
    else
      s := s + Format(' Stop %1.30g',[ds]);
    AStringList.Add(s);

    if ds > 0 then
    begin
      if ds > MaxSingle / 2.0 then
        ds := ds - ((MaxSingle - ds) / 2.0)
      else
        ds := ds / 2.0;
    end
    else
    begin
      if ds < -MaxSingle / 2.0 then
        ds := ds - ((MaxSingle - Abs(ds)) / 2.0)
      else
        ds := ds * 2.0;
    end;
    Inc(cnt)
  until False;
end;


end.

